// File memory.h for CC1B
#ifndef _MEMORY_H_
#define _MEMORY_H_
//include other libraries here
#pragma library 1

/**
 * \file memory.h
 * \brief Library for memory access
 *
 * This library provides functions to access both ram and rom in an easy way. These functions hide the fact\n
 * that ram memory is banked. For optimization reasons, these functions do not perform boundary checks.
 * \n
 */

shadowDef char _SystemRam[256] @0x00;
char _MainTemp @0x0E; //used as temporary placeholder for mainlevel code
char _IsrTemp @0x0D; //used as temporary placeholder for isr code
char _IsrBank @0x0C; //holds rambank for interrupt code

//save register M during interrupt
#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_
char _IsrMode @0xF8; //same as DDRDCOPY, but SX48/52 save M automatically during interrupt in shadow register
#endif

//runtime DDR registers
char DDRACOPY @0xF5;
char DDRBCOPY @0xF6;
#if defined _CHIP_SX28_ || defined _CHIP_SX48_ || defined _CHIP_SX52_
char DDRCCOPY @0xF7;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
char DDRDCOPY @0xF8;
char DDRECOPY @0xF9;
#endif

/**
 * \brief Initialize shadow DDR registers
 *
 * The function DDR_init() initializes the shadow DDR registers to 0xFF. This function should be called\n
 * after clearRAM() at the start of the main() function and before port and pin setup routines.
 *
 * \n
 * How to use:
 * \code
 * DDR_init();
 * \endcode
 * \n
 */
void DDR_init(void)
{
  W = 255;
  DDRACOPY = W;
  DDRBCOPY = W;
#if defined _CHIP_SX28_ || defined _CHIP_SX48_ || defined _CHIP_SX52_
  DDRCCOPY = W;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
  DDRDCOPY = W;
  DDRECOPY = W;
#endif
}

#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_

/**
 * \brief Calculate physical address for _SystemRam[] index
 *
 * The function _SystemRamAddress() converts a logical _SystemRam[] index value into a physical\n
 * ram address. The physical address can be used directly by loading it into FSR and using INDF.
 *
 * \n
 * The mapping from logical index to physical address on the SX18/20/28:
 * - logical -> physical
 * - 0x00-0x0F -> 0x00-0x0F\n
 * - 0x10-0x1F -> 0x10-0x1F\n
 * - 0x20-0x2F -> 0x30-0x3F\n
 * - 0x30-0x3F -> 0x50-0x5F\n
 * - 0x40-0x4F -> 0x70-0x7F\n
 * - 0x50-0x5F -> 0x90-0x9F\n
 * - 0x60-0x6F -> 0xB0-0xBF\n
 * - 0x70-0x7F -> 0xD0-0xDF\n
 * - 0x80-0x8F -> 0xF0-0xFF
 *
 * \n
 * The mapping from logical index to physical address on the SX48/52:
 * - logical -> physical
 * - 0x00-0xFF -> 0x00-0xFF
 *
 * \n
 * How to use:
 * \code
 * addr = _SystemRamAddress(index);
 * \endcode
 *
 * \param W Logical index into _SystemRam[] array (SX18/20/28: 0-143, SX48/52: 0-255).
 * \return Physical ram address.
 * \n
 */
char _SystemRamAddress(char W)
{
  FSR = W;
  W &= 0xF0;
  if (!Zero_) {
    FSR += W;
    FSR -= 0x10;
  }
  return FSR;
}

/**
 * \brief Calculate logical index for  _SystemRam[] from physical address
 *
 * The function _SystemRamIndex() converts a physical address into a logical _SystemRam[] index.\n
 * The logical index value can be used to calculate offsets within arrays.
 *
 * \n
 * The mapping from physical address to logical index on the SX18/20/28:
 * - physical -> logical
 * - 0x00-0x0F -> 0x00-0x0F\n
 * - 0x10-0x1F -> 0x10-0x1F\n
 * - 0x30-0x3F -> 0x20-0x2F\n
 * - 0x50-0x5F -> 0x30-0x3F\n
 * - 0x70-0x7F -> 0x40-0x4F\n
 * - 0x90-0x9F -> 0x50-0x5F\n
 * - 0xB0-0xBF -> 0x60-0x6F\n
 * - 0xD0-0xDF -> 0x70-0x7F\n
 * - 0xF0-0xFF -> 0x80-0x8F
 *
 * \n
 * The mapping from physical address to logical index on the SX48/52:
 * - physical -> logical
 * - 0x00-0xFF -> 0x00-0xFF
 *
 * \n
 * How to use:
 * \code
 * index = _SystemRamAddress(addr);
 * \endcode
 *
 * \param W Physical ram address.
 * \return Logical index into _SystemRam[] array.
 * \n
 */
char _SystemRamIndex(char W)
{
  FSR = W;
  if (FSR.4) {
    FSR.4 = 0;
	W &= 0x0F;
    FSR += W;
	W = FSR >> 1;
    FSR = W;
	FSR += 0x10;
  }
  return FSR;
}

/**
 * \brief Write byte to large array
 *
 * The function LargeArrayWrite() writes a byte to a character array that may cross rambank\n
 * boundaries. No checks are done for address wrapping around 0xFF.
 *
 * \n
 * How to use:
 * \code
 * _MainTemp = value; //value to write must be stored in global ram location
 * LargeArrayWrite(base,index,_MainTemp);
 * \endcode
 *
 * \param base Physical ram address (0x10-0xFF) of array element 0.
 * \param index Logical index into the array.
 * \param value Byte to write.
 * \n
 */
#define LargeArrayWrite(base,index,value) \
{ \
  W = index; \
  FSR = W; \
  W &= 0xF0; \
  FSR += W; \
  W = base; \
  FSR += W; \
  FSR.4 = 1; \
  INDF = value; \
}
  
/**
 * \brief Read byte from large array
 *
 * The function LargeArrayRead() reads a byte from a character array that may cross rambank\n
 * boundaries. No checks are done for address wrapping around 0xFF.
 *
 * \n
 * How to use:
 * \code
 * LargeArrayRead(base,index);
 * value = W;
 * \endcode
 *
 * \param base Physical ram address (0x10-0xFF) of array element 0.
 * \param index Logical index into the array.
 * \return Byte read.
 * \n
 */
#define LargeArrayRead(base,index) \
{ \
  W = index; \
  FSR = W; \
  W &= 0xF0; \
  FSR += W; \
  W = base; \
  FSR += W; \
  FSR.4 = 1; \
  W = INDF; \
}
  
/**
 * \brief Calculate physical address for logical array index
 *
 * The function LargeArrayAddress() converts a logical array index value into a physical ram address.\n
 * The array is specified by the physical address of array element 0.\n
 * The physical address can be used directly by loading it into FSR and using INDF.
 *
 * \n
 * How to use:
 * \code
 * _MainTemp = index; //Logical index into the array.
 * addr = LargeArrayAddress(base);
 * \endcode
 *
 * \param W Base, physical ram address of array element 0.
 * \return Physical ram address of array element index.
 * \n
 */
char LargeArrayAddress(char W)
{
  W = _SystemRamIndex(W);
  W += _MainTemp;
  return _SystemRamAddress(W);
}

/**
 * \brief Calculate logical array index from physical address
 *
 * The function LargeArrayIndex() converts a physical ram address into a logical array index value.\n
 * The array is specified by the physical address of array element 0.\n
 * The logical index can be used directly with the array variable.
 *
 * \n
 * How to use:
 * \code
 * _MainTemp = base; //Physical ram address of array element 0
 * index = LargeArrayIndex(addr);
 * \endcode
 *
 * \param W Addr, physical ram address of array element index.
 * \return Logical index into the array.
 * \n
 */
char LargeArrayIndex(char W)
{
  W = _SystemRamIndex(W);
  _MainTemp ^= W; //swap W and _MainTemp
  W ^= _MainTemp;
  _MainTemp ^= W;
  return _MainTemp - _SystemRamIndex(W);
}

#endif

#if defined _CHIP_SX48_ || defined _CHIP_SX52_

/**
 * \brief Calculate physical address for _SystemRam[] index
 *
 * The function _SystemRamAddress() converts a logical _SystemRam[] index into a physical\n
 * ram address. The physical address can be used directly by loading it into FSR and using INDF.
 *
 * \n
 * The mapping from logical index to physical address on the SX18/20/28:
 * - logical -> physical
 * - 0x00-0x0F -> 0x00-0x0F\n
 * - 0x10-0x1F -> 0x10-0x1F\n
 * - 0x20-0x2F -> 0x30-0x3F\n
 * - 0x30-0x3F -> 0x50-0x5F\n
 * - 0x40-0x4F -> 0x70-0x7F\n
 * - 0x50-0x5F -> 0x90-0x9F\n
 * - 0x60-0x6F -> 0xB0-0xBF\n
 * - 0x70-0x7F -> 0xD0-0xDF\n
 * - 0x80-0x8F -> 0xF0-0xFF
 *
 * \n
 * The mapping from logical index to physical address on the SX48/52:
 * - logical -> physical
 * - 0x00-0xFF -> 0x00-0xFF
 *
 * \n
 * How to use:
 * \code
 * addr = _SystemRamAddress(index);
 * \endcode
 *
 * \param W Logical index into _SystemRam[] array (SX18/20/28: 0-143, SX48/52: 0-255).
 * \return Physical ram address.
 * \n
 */
char _SystemRamAddress(char W)
{
  return W;
}

/**
 * \brief Calculate logical index for  _SystemRam[] from physical address
 *
 * The function _SystemRamIndex() converts a physical address into a logical _SystemRam[] index.\n
 * The logical index value can be used to calculate offsets within arrays.
 *
 * \n
 * The mapping from physical address to logical index on the SX18/20/28:
 * - physical -> logical
 * - 0x00-0x0F -> 0x00-0x0F\n
 * - 0x10-0x1F -> 0x10-0x1F\n
 * - 0x30-0x3F -> 0x20-0x2F\n
 * - 0x50-0x5F -> 0x30-0x3F\n
 * - 0x70-0x7F -> 0x40-0x4F\n
 * - 0x90-0x9F -> 0x50-0x5F\n
 * - 0xB0-0xBF -> 0x60-0x6F\n
 * - 0xD0-0xDF -> 0x70-0x7F\n
 * - 0xF0-0xFF -> 0x80-0x8F
 *
 * \n
 * The mapping from physical address to logical index on the SX48/52:
 * - physical -> logical
 * - 0x00-0xFF -> 0x00-0xFF
 *
 * \n
 * How to use:
 * \code
 * index = _SystemRamIndex(addr);
 * \endcode
 *
 * \param W Physical ram address.
 * \return Logical index into _SystemRam[] array.
 * \n
 */
char _SystemRamIndex(char W)
{
  return W;
}

/**
 * \brief Write byte to large array
 *
 * The function LargeArrayWrite() writes a byte to a character array that may cross rambank\n
 * boundaries. No checks are done for address wrapping around 0xFF.
 *
 * \n
 * How to use:
 * \code
 * _MainTemp = value; //value to write must be stored in global ram location
 * LargeArrayWrite(base,index,_MainTemp);
 * \endcode
 *
 * \param base Physical ram address (0x10-0xFF) of array element 0.
 * \param index Logical index into the array.
 * \param value Byte to write.
 * \n
 */
#define LargeArrayWrite(base,index,value) \
{ \
  W = index; \
  FSR = W; \
  W = base; \
  FSR += W; \
  INDF = value; \
}
  
/**
 * \brief Read byte from large array
 *
 * The function LargeArrayRead() reads a byte from a character array that may cross rambank\n
 * boundaries. No checks are done for address wrapping around 0xFF.
 *
 * \n
 * How to use:
 * \code
 * LargeArrayRead(base,index);
 * value = W;
 * \endcode
 *
 * \param base Physical ram address (0x10-0xFF) of array element 0.
 * \param index Logical index into the array.
 * \return Byte read.
 * \n
 */
#define LargeArrayRead(base,index) \
{ \
  W = index; \
  FSR = W; \
  W = base; \
  FSR += W; \
  W = INDF; \
}
  
/**
 * \brief Calculate physical address for logical array index
 *
 * The function LargeArrayAddress() converts a logical array index into a physical ram address.\n
 * The array is specified by the physical address of array element 0.\n
 * The physical address can be used directly by loading it into FSR and using INDF.
 *
 * \n
 * How to use:
 * \code
 * _MainTemp = index; //Logical index into the array.
 * addr = LargeArrayAddress(base);
 * \endcode
 *
 * \param W Base, physical ram address of array element 0.
 * \return Physical ram address of array element index.
 * \n
 */
char LargeArrayAddress(char W)
{
  return W + _MainTemp;
}

/**
 * \brief Calculate logical array index from physical address
 *
 * The function LargeArrayIndex() converts a physical ram address into a logical array index.\n
 * The array is specified by the physical address of array element 0.\n
 * The logical index can be used directly with the array variable.
 *
 * \n
 * How to use:
 * \code
 * _MainTemp = base; //Physical ram address of array element 0
 * index = LargeArrayIndex(addr);
 * \endcode
 *
 * \param W Addr, physical ram address of array element index.
 * \return Logical index into the array.
 * \n
 */
char LargeArrayIndex(char W)
{
  _MainTemp ^= W; //swap W and _MainTemp
  W ^= _MainTemp;
  _MainTemp ^= W;
  return _MainTemp - W;
}

#endif

/**
 * \brief Read word from rom
 *
 * The function RomWord() reads a word from rom and returns all 12 bits.\n
 * The upper 4 bits are returned in M, the lower 8 bits in W.\n
 * No checks are done wether the specified address is valid.
 *
 * \n
 * How to use:
 * \code
 * RomWord(address);
 * long value;
 * value.low8 = W;
 * value.high8 = MODE; //only if upper 4 bits are needed
 * \endcode
 *
 * \param addr Address of physical rom address (SX18/20/28: 0x000-0x7FF, SX48/52: 0x000-0xFFF).
 * \n
 */
#define RomWord(addr) \
{ \
  W = addr >> 8; \
  MODE = W; \
  W = addr & 0xFF; \
  iread(); \
}

/**
 * \brief Copy bytes from rom to ram
 *
 * The function RomCopy() copies bytes from rom to ram. These bytes are the lower 8 bits\n
 * of the rom words. No checks are done on validity of rom and ram addresses.\n
 * Call this function only from mainline code. For a simular function for TASK subroutines\n
 * see TaskRomCopy() in the Task.h library.
 *
 * \n
 * How to use:
 * \code
 * RomCopy(dest,source,len);
 * \endcode
 *
 * \param dest Physical ram address.
 * \param source Physical rom address (SX18/20/28: 0x000-0x7FF, SX48/52: 0x000-0xFFF).
 * \param len Number of bytes to copy.
 * \n
 */
void RomCopy(char dest, long source, char len)
{
  char i = 0;
  while (len > 0) {
    _MainTemp = i; //set index for LargeArrayAddress()
	char k = LargeArrayAddress(dest);
    RomWord(source);
    _MainTemp = W;
	_SystemRam[k] = _MainTemp;
	source++;
	i++;
	len--;
  }
}

#pragma library 0
#endif
// End of file memory.h
